@@ -1,7 +1,10 @@
 import pythoness
 
-@pythoness.spec("Given a string s and a character c, return the number of times c appears in s.")
-def count_chars(s: str, c: str) -> int:
+@pythoness.spec("Given a string s and a character c, return the number of times c appears in s.",
+                tests = ["count_chars('hello', 'e') == 1",
+                         "count_chars('what is up', 'q') == 0"],
+                max_retries=1)
+def count_chars(s: str, c) -> int:
     ""
 
 print("Running tests.")
